/**
 * This file is part of:
 *    star-ips -- a family of incremental GMRF estimation algorithms.
 *
 * Authors:
 *   Cyril Furtlehner, cyril.furtlehner@inria.fr
 *   Victorin Martin, victorin.martin@gmail.com
 *   Jean-Marc Lasgouttes, jean-marc.lasgouttes@inria.fr
 *
 * This software is distributed under the Creative Commons License
 * BY-NC-SA. The complete license can be found here
 *   http://creativecommons.org/licenses/by-nc-sa/4.0/legalcode
 */

#include "GMRF.h"
#include "Parameters.h"

extern Parameters p;

using namespace std;

GMRF::GMRF(gsl_matrix * PMatrix)
{
	Nn = (int)PMatrix->size1;
	PrecisionMatrix.resize(Nn, Nn);
	if (!gsl_to_std(PMatrix, PrecisionMatrix)) {
		cerr << "-- Error within creation of the ::GMRF:: instance --"
		     << endl;
		cerr << "-- Impossible to convert the precision matrix --" <<
			endl;
		cerr <<
			"-- From this point anything could happen --> Exiting --" <<
			endl;
		exit(1);
	}
	build_neighborhood(p.Nv, p.VERBOSITY);
}

void
GMRF::re_index(int n, int j, int k)
{
	int const n1 = Vois[n][j];
	int const n2 = Vois[n][k];
	int const i1 = ind[n][j];
	int const i2 = ind[n][k];
	Vois[n][j] = n2;
	Vois[n][k] = n1;
	ind[n1][i1] = k;
	ind[n2][i2] = j;
	ind[n][j] = i2;
	ind[n][k] = i1;
}

// Pruning using a threshold.
int GMRF::prune(double th)
{
	int j, c = 0;
	for (int n = 0; n < Nn; n++) {
		for (j = 0; j < nv[n]; j++) {
			if (abs(PrecisionMatrix[n][Vois[n][j]]) < th) {	//Edge not part of the graph
				re_index(n, j, nv[n] - 1);
				nv[n]--;
				na[n]--;
				j--;
				c++;
			}
		}
	}
	int const Narc = std::accumulate(nv.begin(), nv.end(), 0);
	Nc = Nn;
	Na = Nn;
	Km = 1.0 * Narc / Nn;
	return Nc;
}

int GMRF::prune_all()
{
	return GMRF::prune(std::numeric_limits < double >::max());
}

int GMRF::prune_from_Pmatrix (gsl_matrix *A)
{
	int j, c = 0;
	for (int n = 0; n < Nn; n++) {
		for (j = 0; j < nv[n]; j++) {
			if (gsl_matrix_get(A,n,Vois[n][j]) == 0) {	//Edge not part of the graph
				re_index(n, j, nv[n] - 1);
				nv[n]--;
				na[n]--;
				j--;
				c++;
			}
		}
	}
	int const Narc = std::accumulate(nv.begin(), nv.end(), 0);
	Nc = Nn;
	Na = Nn;
	Km = 1.0 * Narc / Nn;
	return Nc;
}

bool GMRF::build_neighborhood(int Nv, int VERBOSITY)
{
	Vois.resize(Nv, Nv);
	ind.resize(Nv, Nv);
	nv.resize(Nv);
	nv0.resize(Nv);
	na.resize(Nv);
	hmu.resize(Nv);
	int ne = 0;
	for (int n1 = 0; n1 < Nv; n1++) {
		for (int n2 = n1 + 1; n2 < Nv; n2++) {
			if (abs(PrecisionMatrix[n1][n2]) > POSITIVE_TH) {
				Vois[n1][nv[n1]] = n2;
				ind[n1][nv[n1]] = nv[n2];
				Vois[n2][nv[n2]] = n1;
				ind[n2][nv[n2]] = nv[n1];
				nv[n1]++;
				nv[n2]++;
				ne++;
			}
		}
	}
	for (int i = 0; i < Nv; i++) {
		na[i] = nv[i];
		nv0[i] = nv[i];
	}
	if (VERBOSITY > 1) {
		cout << "-- GMRF Model created -- Number of variables: " << Nv
			 << " -- Mean connectivity: " << 2.0 * ne / Nv
			 << " -- " << endl;
	}
	return true;
}

void GMRF::set_XX_YY_Mask(int NvX,int NvY)         //blacklist all inter links between X and Y
{
	cout << "Set XX-YY mask\n"; 
	int N = NvX+NvY;
	for (int n1=0;n1<N;n1++)
		nv0[n1] = N-1;
	for (int n1=0;n1<NvX;n1++)
	{
		for (int j=0;j<nv0[n1];j++)
		{
			int n2 = Vois[n1][j];
			int ii = ind[n1][j];
			if (n2 >= NvX)
			{
				re_index(n1, j, nv0[n1] - 1);
				nv0[n1]--;
				re_index(n2, ii, nv0[n2] - 1);
				nv0[n2]--;
				j--;
			}
		}
	}
}

void GMRF::set_XY_Mask(int NvX,int NvY)           //blacklist all intern links in X and Y
{
	cout << "Set XY mask\n"; 
	int N = NvX+NvY;
	for (int n1=0;n1<N;n1++)
		nv0[n1] = N-1;
	for (int n1=0;n1<N;n1++)
	{
		for (int j=nv[n1];j<nv0[n1];j++)
		{
			int n2 = Vois[n1][j];
			int ii = ind[n1][j];
			if ((n1< NvX && n2 < NvX) || (n1 >= NvX && n2 >= NvX) )
			{
				re_index(n1, j, nv0[n1] - 1);
				nv0[n1]--;
				re_index(n2, ii, nv0[n2] - 1);
				nv0[n2]--;
				j--;
			}
		}
	}
}
