/**
 * This file is part of:
 *    star-ips -- a family of incremental GMRF estimation algorithms.
 *
 * Authors:
 *   Cyril Furtlehner, cyril.furtlehner@inria.fr
 *   Victorin Martin, victorin.martin@gmail.com
 *   Jean-Marc Lasgouttes, jean-marc.lasgouttes@inria.fr
 *
 * This software is distributed under the Creative Commons License
 * BY-NC-SA. The complete license can be found here
 *   http://creativecommons.org/licenses/by-nc-sa/4.0/legalcode
 */

#include "Tools.h"

using namespace std;

// Finding (only real) Roots of degree two polynomial.
p2_roots find_roots(double a, double b, double c)
{
	struct p2_roots roots = { -1e8, 1e8, false };
	if (fabs(b / a) > 1e10) {	//A degree one polynomial.
		if (fabs(b) < 1e-20)
			return roots;
		roots.are_roots_real = true;
		roots.r1 = -c / b;
		roots.r2 = roots.r1;
		return roots;
	}
	double const delta = b * b - 4 * a * c;
	if (delta < 0) {
		return roots;
	}
	double const sqrtdelta = sqrt(delta);
	roots.are_roots_real = true;
	roots.r1 = (-b - sqrtdelta) / (2 * a);
	roots.r2 = (-b + sqrtdelta) / (2 * a);
	return roots;
}

//Eigenvalue decomposition.
void eigenval(gsl_matrix * M, gsl_vector * lp)
{
	int const n = M->size1;
	gsl_matrix *copieM;
	copieM = gsl_matrix_alloc(n, n);
	gsl_matrix_memcpy(copieM, M);
	gsl_eigen_symm_workspace *W = gsl_eigen_symm_alloc(n);
	gsl_eigen_symm(copieM, lp, W);
	gsl_eigen_symm_free(W);
	gsl_matrix_free(copieM);
}

double scalar_product(const gslmm::vector & v1, const gslmm::vector & v2)
{
	double sp = 0;
	gsl_blas_ddot(v1, v2, &sp);
	return sp;
}

//Matrix inversion.
gslmm::matrix inverse(const gslmm::matrix & M)
{
	int n = M.size1(),s;
	gslmm::matrix copieM(n, n), IM(n, n);
	if (M.size1() != M.size2()) {
		cerr << "error: Trying to invert a non-square matrix!" << endl;
		IM.resize(0,0);
		return IM;
	}

	gsl_permutation *P;
	P = gsl_permutation_alloc(n);

	copieM = M;
	gsl_linalg_LU_decomp(copieM, P, &s);
	gsl_linalg_LU_invert(copieM, P, IM);

	gsl_permutation_free(P);

	return IM;
}

double lndet(const gslmm::matrix & M)
{
	int n = M.size1(), s;
	if (M.size1() != M.size2()) {
		cerr << "Trying to compute the determinant of a non-square matrix -- will return 0"
		     << endl;
		return 0;
	}
	gslmm::matrix copieM(n,n);
	gsl_permutation *P;
	P = gsl_permutation_alloc(n);

	copieM = M;
	gsl_linalg_LU_decomp(copieM, P, &s);

	double x = gsl_linalg_LU_lndet(copieM);
	gsl_permutation_free(P);

	return x;
}

double trace(const gslmm::matrix & M)
{
	double x = 0;
	for (size_t i = 0; i < M.size1(); i++)
		x += M(i, i);
	return x;
}

///**** Read a (square) matrix from a file (M should be of the right size) **************
bool import_matrix(gsl_matrix * M, string const & s)
{
	int const n = M->size1, m = M->size2;
	double mij;
	if (n != m) {
		cerr << "Not a square matrix in import_matrix" << endl;
		return false;
	}
	ifstream from_file;
	from_file.open(s.c_str());
	if (!from_file.is_open())
		return false;
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {
			from_file >> mij;
			gsl_matrix_set(M, i, j, mij);
		}
	}
	return true;
}

bool export_matrix(gsl_matrix const * M, string const & s)
{
	ofstream ofs;
	ofs.open(s.c_str());
	if (!ofs.is_open())
		return false;

	for (size_t i = 0; i < M->size1; i++) {
		for (size_t j = 0; j < M->size2; j++)
			ofs << gsl_matrix_get(M, i, j) << " ";
		ofs << endl;
	}
	ofs.close();
	return true;
}

void compare_graph(gsl_matrix * MRef, gsl_matrix * M, ostream& os)
{
	size_t false_pos = 0;
	size_t false_neg = 0;
	size_t nedges = 0;
	for (size_t i = 0; i < MRef->size1; i++) {
		for (size_t j = i + 1; j < MRef->size2; j++) {
			bool isInRef = fabs(gsl_matrix_get(MRef, i, j)) > 1e-9;
			bool isInM = fabs(gsl_matrix_get(M, i, j)) > 1e-9;
			if (isInM)
				nedges++;
			if (isInM != isInRef) {
				if (isInRef)
					false_neg++;
				else
					false_pos++;
			}
		}
	}
	double k = (2.0*nedges)/(1.0*M->size1);
	os << k << "\t" << false_pos << "\t"
	   << false_neg << "\t" << nedges << endl;
}

bool gsl_to_std(gsl_matrix * Mgsl, vector2 < double >&Mstd)
{
	if ((Mgsl->size1 != Mstd.size1())
	    || (Mgsl->size2 != Mstd.size2())) {
		cerr << "Conversion between matrices of different sizes"
		     << endl;
		return false;
	}
	for (size_t i = 0; i < Mgsl->size1; i++) {
		for (size_t j = 0; j < Mgsl->size2; j++) {
			Mstd[i][j] = gsl_matrix_get(Mgsl, i, j);
		}
	}
	return true;
}
