// -*- C++ -*-
/**
 * This file is part of:
 *    star-ips -- a family of incremental GMRF estimation algorithms.
 *
 * Authors:
 *   Cyril Furtlehner, cyril.furtlehner@inria.fr
 *   Victorin Martin, victorin.martin@gmail.com
 *   Jean-Marc Lasgouttes, jean-marc.lasgouttes@inria.fr
 *
 * This software is distributed under the Creative Commons License
 * BY-NC-SA. The complete license can be found here
 *   http://creativecommons.org/licenses/by-nc-sa/4.0/legalcode
 */

#ifndef TOOLS_H
#define TOOLS_H

#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <vector>
#include <sstream>
#include <algorithm>
#include <cstring>
#include <numeric>
#include <limits>
#include <queue>
#include <gsl/gsl_math.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>
#include "gsl--/matrix"
#include "gsl--/vector"

#include "simd_math_prims.h"
#define log logapprox

#define POSITIVE_TH 1e-6

//GSL matrix and vector product.
inline gslmm::matrix
prod_matrix (const gslmm::matrix & Mleft, const gslmm::matrix & Mright)
{
	int n = Mleft.size1();
	gslmm::matrix C(n,n);
	gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1, Mleft, Mright, 0, C);
	return C;
}

inline gslmm::vector
prod_matrix (const gslmm::matrix & M, const gslmm::vector & U)
{
	int n = M.size1();
	gslmm::vector V(n);
	gsl_blas_dgemv (CblasNoTrans, 1.0, M, U, 0, V);
	return V;
}

//Dimension 2 vector.
template < typename T > class vector2
{
public:
	vector2 () {
	};

	vector2 (int rows, int cols) : data_(rows, std::vector < T > (cols)) {}

	inline std::vector < T > &operator[] (int i) {
		return data_[i];
	}

	inline const std::vector < T > &operator[] (int i) const {
		return data_[i];
	}

	size_t size1() const {
		return data_.size();
	}

	size_t size2() const {
		return data_.empty() ? 0 : data_.front().size();
	}

	void resize (int rows, int cols) {
		data_.resize (rows);
		for (int i = 0; i < rows; ++i)
			data_[i].resize (cols);
	}

	void resize (int rows, int cols, T value) {
		data_.resize (rows);
		for (int i = 0; i < rows; ++i)
			data_[i].resize (cols, value);
	}

private:
	std::vector < std::vector < T > >data_;
};

///Structure for roots of a degree two polynomial.
struct p2_roots {
	double r1, r2;
	bool are_roots_real;

	// Is there at least one root between a and b?
	bool in_range (double a, double b) {
		if (!are_roots_real)
			return false;
		return (r1 < b && r1 > a) || (r2 < b && r2 > a);
	}
};

// Compute roots of a degree 2 polynomial
p2_roots find_roots (double, double, double);



///
void eigenval (gsl_matrix *, gsl_vector *);
gslmm::matrix inverse (const gslmm::matrix &);
double scalar_product(const gslmm::vector &, const gslmm::vector &);
double lndet (const gslmm::matrix &);
double trace (const gslmm::matrix &);
bool export_matrix (gsl_matrix const *, std::string const &);
bool import_matrix (gsl_matrix *, std::string const &);
bool gsl_to_std (gsl_matrix *, vector2 < double >&);
void compare_graph (gsl_matrix *, gsl_matrix *, std::ostream&);

#endif // TOOLS_H
